/*
  Copyright 2003 John Knottenbelt
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
*/

#include "config.h"
#include "dvbd.h"
#include "utils.h"
#include "stringutil.h"
#include "select.h"
#include "clientconnection.h"

#include <string>
#include <iostream>
#include <time.h>
#include <getopt.h>
#include <cerrno>
#include <fstream>

void usage(const char *progname);

int main(int argc, char *argv[])
{
  // improve iostream performance
  std::ios::sync_with_stdio(false);

  std::string socketFile = DEFAULT_SOCKET_FILE;
  int priority = 10;
  bool listConnections = false;
  bool listSchedule = false;
  bool listNextJob = false;

  while (true) {
    static struct option long_options[] = {
      {"priority", 1, 0, 'p'},
      {"socket", 1, 0, 's'},
      {"connections", 0, 0, 'c'},
      {"schedule", 0, 0, 'h'},
      {"nextjob", 0, 0, 'n'},
      {0, 0, 0, 0}
    };
 
    int option_index = 0;
    int c = getopt_long(argc, argv, "chnp:s:", long_options, &option_index);
    if (c == -1)
      break;
    switch (c) {
    case 'p':
      priority = toInt(optarg);
      break;
    case 's':
      socketFile = optarg;
      break;
    case 'c':
      listConnections = true;
      break;
    case 'h':
      listSchedule = true;
      break;
    case 'n':
      listNextJob = true;
      break;

    default:
      usage(argv[0]);
      return 1;
    }
  }

  ClientConnection client(socketFile);
  if (!client) {
    std::cerr << "Failed to connect to server on socket " << socketFile << std::endl;
    return 1;
  }

  // List current connections
  if (listConnections) {
    StringList connections;
    if (!client.listConnections(connections)) {
      std::cerr << "Failed to list connections on server.\n";
      return 1;
    }

    std::cout << "Current connections are:\n";
    for (StringList::iterator i = connections.begin(); i != connections.end(); i++)
      std::cout << *i << "\n";
    return 0;
  }

  if (listSchedule) {
    StringList schedule;
    if (!client.listSchedule(schedule)) {
      std::cerr << "Failed to list schedules on server.\n";
      return 1;
    }

    std::cout << "Schedule is:\n";
    for (StringList::iterator i = schedule.begin(); i != schedule.end(); i++)
      std::cout << *i << "\n";
    return 0;
  }

  if (listNextJob) {
    time_t nextJobTime;
    if (!client.getNextJobTime(nextJobTime)) {
      std::cerr << "No next job!\n";
      return 1;
    }
    else {
      std::cout << stripTrailingWS(ctime(&nextJobTime)) << "\n";
      return 0;
    }
  }

  if (argc - optind != 1) {
    if (argc - optind != 0)
      std::cerr << "Incorrect number of arguments passed (expected 1, got " 
		<< argc - optind << ")\n\n";

    usage(argv[0]);
    return 1;
  }

  // Parse the command line arguments
  std::string type(argv[optind++]);

  StringList channels;
  if (!client.getChannels(type, priority, channels)) {
    std::cerr << "Failed to get channels from server.\n";
    return 1;
  }

  std::cout << type << " channels available to tune at priority " << priority << "\n";
  for (StringList::iterator i = channels.begin(); i != channels.end(); i++)
    std::cout << *i << "\n";

  return 0;
}

void usage(const char *progname) {
  std::cerr << "Usage: " << progname << " [OPTION] TYPE\n"
	    << "List DVB channels of type TYPE available for a given priority\n"
	    << "\n"
	    << "TYPE can be dvb-t, dvb-s or dvb-c.\n"
	    << "\n"
	    << "  -p, --priority         Priority of stream (default is 10\n"
	    << "                         higher priorities may dislodge lower\n"
	    << "                         ones).\n"
	    << "  -s, --socket           Specify socket to connect to\n"
	    << "                         (default is " << DEFAULT_SOCKET_FILE << ")\n"
	    << "  -c, --connections      List current connections\n"
	    << "                         (no need to give TYPE)\n"
	    << "  -h, --schedule         List upcoming recordings\n"
	    << "                         (no need to give TYPE)\n"
	    << "  -n, --nextjob          Show date of next job\n";

}
