/*
  Copyright 2003 John Knottenbelt
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
*/

#include "config.h"
#include "dvbd.h"
#include "tunermanager.h"
#include "tuner.h"
#include "tuneparams.h"
#include "configfile.h"
#include "stringutil.h"
#include <iostream>
#include <cassert>
#include <set>
#include <vector>

TunerManager::~TunerManager()
{
  // Destroy the tuners
  for (TunerList::iterator i = tuners.begin(); i != tuners.end(); i++)
    delete *i;
}

void TunerManager::parseCommandLine(int argc, char *argv[])
{
  // Read in the type / tuner / channels triples
  for (int x = 0; x < argc; x+=3) 
    add(argv[x], argv[x+1], argv[x+2]);
}

bool TunerManager::parseConfig(ConfigFile &c)
{
  int numcards = toInt(c.get("numcards"));

  for (int i = 0; i < numcards; i++) {
    std::string card = c.get("card" + toString(i));
    std::vector<std::string> words;

    splitString(card, ':', words);
    if (words.size() != 3) {
      std::cerr << "Error parsing card definition. Expected 3 segments, got "
		<< words.size() << "\n"
		<< card 
		<< std::endl;
      return false;
    }
    add(words[0], words[1], words[2]);
  }

  return true;
}

bool TunerManager::add( const std::string &type, 
			const std::string &device, 
			const std::string &channels )
{
  ChannelMap *cm = new ChannelMap;

  if (!readChannelsConf(type, channels, *cm)) {
    std::cerr << "Fatal error: failed to read " << channels << " file\n";
    return false;
  }
    
  Tuner *t = new Tuner(device);

  tuners.push_back(t);
  tunerChannelMap[t] = cm;
    
  TunerList *&typedTuners = typeTunerMap[type];
  if (typedTuners == NULL)
    typedTuners = new TunerList;

  typedTuners->push_back(t);  
  return true;
}

bool TunerManager::validType(const std::string &type) const
{
  return typeTunerMap.find(type) != typeTunerMap.end();
}

bool TunerManager::validChannel(const std::string &type, const std::string &channel) const
{
  TypeTunerMap::const_iterator li = typeTunerMap.find(type);
  if (li == typeTunerMap.end())
    return false;

  const TunerList &l = *li->second;

  for (TunerList::const_iterator i = l.begin(); i != l.end(); i++) {
    Tuner *t = *i;

    TunerChannelMap::const_iterator j = tunerChannelMap.find(t);
    assert(j != tunerChannelMap.end());

    ChannelMap *cm = j->second;

    ChannelMap::const_iterator k = cm->find(channel);
    if (k != cm->end())
      return true;
  }

  // No tuner found to tune this channel
  return false;
}

void TunerManager::getTunableChannels(const std::string &type, int priority, StringList &channels) const
{
  TypeTunerMap::const_iterator li = typeTunerMap.find(type);
  if (li == typeTunerMap.end())
    return;

  const TunerList &l = *li->second;
  std::set<std::string> history;

  for (TunerList::const_iterator i = l.begin(); i != l.end(); i++) {
    Tuner *t = *i;

    TunerChannelMap::const_iterator j = tunerChannelMap.find(t);
    assert(j != tunerChannelMap.end());

    ChannelMap *cm = j->second;

    for (ChannelMap::const_iterator k = cm->begin();
	 k != cm->end(); k++) {
      const std::string &channel = k->first;

      if (history.find(channel) == history.end() &&
	  canGetDemuxers(type, channel, priority)) {
	channels.push_back(channel);
	history.insert(channel);
      }
    }
  }
}

bool TunerManager::canGetDemuxers(const std::string &type, const std::string &channel,
				  int priority) const
{
  // Without actually allocating a demuxer 
  // see if it is possible to tune the channel.

  TypeTunerMap::const_iterator li = typeTunerMap.find(type);
  assert(li != typeTunerMap.end());

  const TunerList &l = *li->second;

  for (TunerList::const_iterator i = l.begin(); i != l.end(); i++) {
    Tuner *t = *i;
    ChannelMap *cm = tunerChannelMap.find(t)->second;
    ChannelMap::iterator j = cm->find(channel);

    if (j != cm->end()) {
      const TuneParams *c = j->second;

      // Tuner with no subscribers
      if (t->getNumSubscribers() == 0)
	return true;

      if (!t->getTuneParams()->sameAsExceptPIDs(*c)) {
	// Tuner on a different frequency with subscribers all of
	// lesser priority
	if (t->getPriority() < priority)
	  return true;
	else 
	  continue;
      }

      Source *audio = t->getDemuxer(c->getAudioPID(), false, Demuxer::Audio);
      Source *video = t->getDemuxer(c->getVideoPID(), false, Demuxer::Video);

      // Check to see if perhaps someone else has already
      // tuned the audio or video channels (this should
      // not result in a new subscription to the tuner).
      int additionalRequired = 0;

      if (audio && audio->getNumSubscribers() == 0)
	additionalRequired++;

      if (video && video->getNumSubscribers() == 0)
	additionalRequired++;

      // Check if the tuner has the PIDS available
      // (ignoring subscriptions of lower priority)
      if (t->getNumSubscribers(priority) + additionalRequired < t->getMaxNumSubscribers()) 
	return true;
    }
  }

  return false;
}

void TunerManager::getDemuxers(const std::string &type, const std::string &channel,
			       bool convertToPS, SourceList &demuxers, int priority) 
{
  TypeTunerMap::iterator li = typeTunerMap.find(type);
  if (li == typeTunerMap.end()) {
    std::cerr << "Fatal error: Could not find tuner for type " 
	      << type << "(channel " << channel << ")" << std::endl;
    return;
  }

  const TunerList &l = *li->second;

  demuxers.clear();

  // Stratgey:
  // Try to allocate to a tuner already in use
  // Try to allocate to a free tuner

 tryAgain:
  for (TunerList::const_iterator i = l.begin(); i != l.end(); i++) {
    Tuner *t = *i;
    ChannelMap *cm = tunerChannelMap.find(t)->second;
    ChannelMap::iterator j = cm->find(channel);

    if (j != cm->end()) {
      const TuneParams *c = j->second;

      if (t->getNumSubscribers() == 0)
	t->setTuneParams(c);

      if (!t->getTuneParams()->sameAsExceptPIDs(*c)) 
	continue;

      Source *audio = t->getDemuxer(c->getAudioPID(), convertToPS, Demuxer::Audio);
      Source *video = t->getDemuxer(c->getVideoPID(), convertToPS, Demuxer::Video);

      // Check to see if perhaps someone else has already
      // tuned the audio or video channels (this should
      // not result in a new subscription to the tuner).
      int additionalRequired = 0;

      if (audio && audio->getNumSubscribers() == 0)
	additionalRequired++;

      if (video && video->getNumSubscribers() == 0)
	additionalRequired++;

      // Check if the tuner has the PIDS available
      if (t->getNumSubscribers() + additionalRequired >= t->getMaxNumSubscribers()) 
	continue;

      if (audio) 
	demuxers.push_back(audio);

      if (video)
	demuxers.push_back(video);

      return;
    }
  }

  // Try to free up a tuner/demuxers based on priority
  for (TunerList::const_iterator i = l.begin(); i != l.end(); i++) {
    Tuner *t = *i;
    ChannelMap *cm = tunerChannelMap.find(t)->second;
    ChannelMap::iterator j = cm->find(channel);

    if (j != cm->end()) {
      const TuneParams *c = j->second;

      if (t->getTuneParams()->sameAsExceptPIDs(*c)) {
	// We need to kick off a demuxer, if possible
	if (t->Source::unsubscribe(priority) > 0)
	  goto tryAgain;
      }
      else {
	// We need to kick off all subscribers,
	// if possible
	if (priority > t->getPriority()) {
	  t->unsubscribeAll();
	  goto tryAgain;
	}
      }
    }
  }

  // We failed :(
}

void TunerManager::addSelectFDs(Select &s) const
{
  for (TunerList::const_iterator i = tuners.begin(); i != tuners.end(); i++)
    (*i)->addSelectFDs(s);
}

bool TunerManager::isReady(const Select &s) const
{
  for (TunerList::const_iterator i = tuners.begin(); i != tuners.end(); i++)
    if ((*i)->isReady(s))
      return true;
  return false;
}

void TunerManager::processReady(const Select &s)
{
  for (TunerList::iterator i = tuners.begin(); i != tuners.end(); i++)
    if ((*i)->isReady(s))
      (*i)->process();
}

